// ヘッダーとインデックステーブルの取得用
var indexOffset;
// 編集後のデータのオブジェクト
var editEntries;
// ファイル名
var sfoFileName;
// モーダル内部が未作成かどうか
var modalNotCreated = true;

// タイプ判別用
var PSF_TYPE_UTF8S = 0;
var PSF_TYPE_UTF8 = 2;
var PSF_TYPE_INT32 = 4;


function parseSfo(sfoArrayBuffer, def){
    // グローバル変数の初期化
    indexOffset = 0;
    // ヘッダーを取得する必要があるので20バイトは必要
    if(sfoArrayBuffer.byteLength >= 20){
        // 取得したsfoのデータを8bit配列に変換
        var u8Array = new Uint8Array(sfoArrayBuffer);
        // ヘッダーを取得
        var sfoHeader = {
            "magic":readUint32(u8Array),
            "version":readUint32(u8Array),
            "keyofs":readUint32(u8Array),
            "valofs":readUint32(u8Array),
            "count":readUint32(u8Array)
        }

        //\x00PSF
        if(sfoHeader.magic == "0x46535000"){
            // インデックステーブルのデータを取得
            var sfoIndexes = [];
            for(var i=0; i<sfoHeader.count; i++){
				var sfoIndex = {
					"nameofs":readUint16(u8Array),
					"alignment":readUint8(u8Array),
					"type":readUint8(u8Array),
					"valsize":readUint32(u8Array),
					"totalsize":readUint32(u8Array),
					"dataofs":readUint32(u8Array)
				}
				sfoIndexes = sfoIndexes.concat(sfoIndex);
			}

            // キーテーブルとデータテーブルからデータを取得
            var sfoEntries = {};
            for(var i=0; i<sfoIndexes.length; i++){
                var keyOffset = sfoHeader.keyofs + sfoIndexes[i].nameofs;
                var keySize = keyOffset + readKeyBytes(u8Array, keyOffset);
                var keyName = readString(u8Array, keyOffset, keySize);
                var valOffset = sfoHeader.valofs + sfoIndexes[i].dataofs;
                var valSize = valOffset + sfoIndexes[i].valsize;
                // パディング数を算出
                var valPadding = 0;
                if(i < sfoIndexes.length-1){
                    var nextValOffset = sfoHeader.valofs + sfoIndexes[i+1].dataofs;
                    valPadding = nextValOffset - valSize;
                }else{
                    var endOffset = u8Array.length;
                    valPadding = endOffset - valSize;
                }
                // 各Valueの取得
                var value = "";
                switch(sfoIndexes[i].type){
                    case PSF_TYPE_UTF8:
                        value = readString(u8Array, valOffset, valSize);
                        break;
                    case PSF_TYPE_INT32:
                        value = (function(){
                            var Ary = u8Array.slice(valOffset, valOffset+4);
                            return new Uint32Array(Ary.buffer)[0];
                        }());
                        break;
                    default:
                        value = readHex(u8Array, valOffset, valSize);
                        break;
                }
                sfoEntries[keyName] = {
                    "value": value,
                    "length": sfoIndexes[i].valsize,
                    "maxLength": sfoIndexes[i].totalsize,
                    "type": sfoIndexes[i].type,
                    "valpadding": valPadding
                }
            }

            //console.log(sfoHeader);
            //console.log(sfoIndexes);
            //console.log(sfoEntries);

            def.resolve(sfoEntries);

        }else{
            def.reject("This file is not SFO.");
        }
    }else{
        def.reject("This file is less than 20 bytes. Cannot get the header.")
    }
}

function readUint8(u8Array){
    var u8 = u8Array[indexOffset];
    indexOffset++;
    return u8;
}

function readUint16(u8Array){
    var ary = u8Array.slice(indexOffset, indexOffset+2);
    var u16 = new Uint16Array(ary.buffer)[0];
    indexOffset += 2;
    return u16;
}

function readUint32(u8Array){
    var ary = u8Array.slice(indexOffset, indexOffset+4);
    var u32 = new Uint32Array(ary.buffer)[0];
    indexOffset += 4;
    return u32;
}

function readKeyBytes(u8Array, keyOffset){
    for(var keySize=0; ; keySize++){
        u8 = u8Array[keyOffset+keySize];
        if(u8 == 0x00){
            return keySize;
        }
    }
}

function readString(u8Array, strOffset, strSize){
    var ary = u8Array.slice(strOffset, strSize);
    return new TextDecoder("utf-8").decode(ary);
}

function readHex(u8Array, hexOffset, hexSize){
    var ary = u8Array.slice(hexOffset, hexSize);
    var hexAry = [];
    for(var i=0; i<ary.length; i++){
        var hex = ary[i].toString(16);
        if(hex.length==1){
            hex = "0" + hex;
        }
        hexAry.push(hex);
    }
    return hexAry.join("");
}

function stringToBytes(str){
    var u8Array = new TextEncoder().encode(str);
    return Array.from(u8Array);
}

function uintToBytes(uintValue, byteLength){
    var valueAryBuffer;
    switch(byteLength){
        case 4:
            valueAryBuffer = new Uint32Array([uintValue]).buffer;
            break;
        case 2:
            valueAryBuffer = new Uint16Array([uintValue]).buffer;
            break;
        case 1:
            valueAryBuffer = new Uint8Array([uintValue]).buffer;
            break;
    }
    const u8Array = new Uint8Array(valueAryBuffer);
    return Array.from(u8Array);
}

function hexStringToBytes(hexString){
    var ary = hexString.split("");
    var half = Math.round(ary.length/2);
    var bytes = [];
    for(var i=0; i<half; i++){
        if(i==half-1 && ary.length==1){
            var hex = "0" + ary.shift();
            hex = "0x" + hex;
            bytes.push(parseInt(hex));
        }else{
            var hex = [ary.shift(), ary.shift()].join("");
            hex = "0x" + hex;
            bytes.push(parseInt(hex));
        }
    }
    return bytes;
}

function createSfo(sfoEntries){
    console.log(sfoEntries);
    /* キーテーブルとデータテーブルの作成 */
    const sfoKeyNames = Object.keys(sfoEntries);
    var sfoKeyBytesAry = [];
    var sfoDataBytesAry = [];
    for(var i=0; i<sfoKeyNames.length; i++){
        const keyName = sfoKeyNames[i];
        // キーテーブル
        var keyBytes = stringToBytes(keyName);
        keyBytes.push(0);
        sfoKeyBytesAry.push(keyBytes);
        // データテーブル
        const dataValue = sfoEntries[keyName]["value"];
        const dataType = sfoEntries[keyName]["type"];
        var dataPadding = sfoEntries[keyName]["valpadding"];
        const dataLength = sfoEntries[keyName]["length"];
        var dataBytes;
        switch(dataType){
            case PSF_TYPE_UTF8:
                dataBytes = stringToBytes(dataValue);
                break;
            case PSF_TYPE_INT32:
                dataBytes = uintToBytes(dataValue, 4);
                break;
            default:
                dataBytes = hexStringToBytes(dataValue);
                break;
        }
        // データテーブルにパディングを付与
        var dataMaxLength = sfoEntries[keyName]["maxLength"];
        var adjustPadding = dataMaxLength - dataLength;
        if(dataPadding != adjustPadding && adjustPadding > 0){
            // Valueが更新されていた場合は付与する数を調整
            adjustPadding = dataPadding - adjustPadding;
            dataPadding = dataPadding - adjustPadding;
        }
        for(var x=0; x<dataPadding; x++){
            dataBytes.push(0);
        }
        sfoDataBytesAry.push(dataBytes);
    }
    // キーテーブルにパディングを追加
    var sfoKeyBytesAryInternalLength = 0
    for(var i=0; i<sfoKeyBytesAry.length; i++){
        sfoKeyBytesAryInternalLength += sfoKeyBytesAry[i].length;
    }
    var keyTablePadding = sfoKeyBytesAryInternalLength % 4;
    if(keyTablePadding != 0){
        // キーテーブルはサイズを4バイトの乗算で揃える必要があるため0を追加
        keyTablePadding = 4 - keyTablePadding;
        for(var i=0; i<keyTablePadding; i++){
            sfoKeyBytesAry[sfoKeyBytesAry.length-1].push(0)
        }
    }

    /* ヘッダーの作成 */
    const magicBytes = [0, 80, 83, 70];
    const versionBytes = [1, 1, 0, 0];
    // ヘッダーは20バイト、インデックステーブルの各エントリは16バイト
    const keyTableOffset = 20 + (sfoKeyNames.length * 16);
    const dataTableOffset = keyTableOffset + sfoKeyBytesAryInternalLength + keyTablePadding;
    const keytableBytes = uintToBytes(keyTableOffset, 4);
    const dataTableBytes = uintToBytes(dataTableOffset, 4);
    const numberOfEntries = uintToBytes(sfoKeyNames.length, 4);
    var sfoHeaderBytes = [];
    Array.prototype.push.apply(sfoHeaderBytes, magicBytes);
    Array.prototype.push.apply(sfoHeaderBytes, versionBytes);
    Array.prototype.push.apply(sfoHeaderBytes, keytableBytes);
    Array.prototype.push.apply(sfoHeaderBytes, dataTableBytes);
    Array.prototype.push.apply(sfoHeaderBytes, numberOfEntries);

    /* インデックステーブルの作成 */
    var sfoIndexTableBytes = [];
    var keyOffset = 0;
    var dataOffset = 0;
    for(var i=0; i<sfoKeyBytesAry.length; i++){
        if(i>0){
            keyOffset += sfoKeyBytesAry[i-1].length;
            dataOffset += sfoDataBytesAry[i-1].length;
        }
        const keyOffsetBytes = uintToBytes(keyOffset, 2)
        //keyOffsetBytes.push(4);
        const keyName = sfoKeyNames[i];
        const dataType = sfoEntries[keyName]["type"];
        const dataTypeBytes = [4, uintToBytes(dataType, 1)[0]];
        const dataLength = sfoEntries[keyName]["length"];
        const dataLengthBytes = uintToBytes(dataLength, 4);
        const dataMaxLength = sfoEntries[keyName]["maxLength"];
        const dataMaxLengthBytes = uintToBytes(dataMaxLength, 4);
        const dataOffsetBytes = uintToBytes(dataOffset, 4);
        Array.prototype.push.apply(sfoIndexTableBytes, keyOffsetBytes);
        Array.prototype.push.apply(sfoIndexTableBytes, dataTypeBytes);
        Array.prototype.push.apply(sfoIndexTableBytes, dataLengthBytes);
        Array.prototype.push.apply(sfoIndexTableBytes, dataMaxLengthBytes);
        Array.prototype.push.apply(sfoIndexTableBytes, dataOffsetBytes);
    }

    /* キーテーブル、データテーブルを収めた多次元配列の各配列を結合 */
    var sfoKeyBytes = [];
    for(var i=0; i<sfoKeyBytesAry.length; i++){
        sfoKeyBytes = sfoKeyBytes.concat(sfoKeyBytesAry[i]);
    }
    var sfoDataBytes = [];
    for(var i=0; i<sfoDataBytesAry.length; i++){
        sfoDataBytes = sfoDataBytes.concat(sfoDataBytesAry[i]);
    }
    
    /* sfoの作成 */
    var sfoBytes = [];
    Array.prototype.push.apply(sfoBytes, sfoHeaderBytes);
    Array.prototype.push.apply(sfoBytes, sfoIndexTableBytes);
    Array.prototype.push.apply(sfoBytes, sfoKeyBytes);
    Array.prototype.push.apply(sfoBytes, sfoDataBytes);
    let sfo = new Uint8Array;
    sfo = Uint8Array.from(sfoBytes);

    return sfo;
}

function displaySfo(sfoEntries, FileName, table){
    $(table).show();
    var sfoKeyNames = Object.keys(sfoEntries);
    for(var i=0; i<sfoKeyNames.length; i++){
        var keyName = sfoKeyNames[i];
        var dataValue = sfoEntries[keyName]["value"];
        var dataLength = sfoEntries[keyName]["length"];
        var dataMaxLength = sfoEntries[keyName]["maxLength"];
        dataLength = "<span class='nume'>" + dataLength + "</span>";
        dataMaxLength = "<span class='deno'>" + dataMaxLength + "</span>" ;
        dataLength = dataLength + "/" + dataMaxLength;
        var dataType = sfoEntries[keyName]["type"];
        var keyElem = "<th scope='row' title='" + keyName + "'>" + keyName + "</th>";
        var valueElem = "<td class='sfo-value'>" + dataValue + "</td>";
        var lengthElem = "<td class='sfo-length'>" + dataLength + "</td>";
        var typeElem = "<td class='sfo-type'>" + dataType + "</td>";
        var trElem = "<tr data-key='"+keyName+"'>" + keyElem + valueElem + lengthElem + typeElem + "</tr>";
        $(table).find("tbody").append(trElem);
    }
    sfoFileName = FileName;
    $("#file-related-wrapper").show();
    $("#edit-file-name").val(sfoFileName);
}


function editSfo(sfoEntries){
    /* テーブルと編集エリアを表示 */
    $("#table-wrapper").show();
    $("#editor-wrapper").show();
    
    /* オリジナルのオブジェクトを直接触らないようにする */
    editEntries = $.extend(true, {}, sfoEntries);

    var keyName;
    var dataValue;
    var dataLength;
    var dataMaxLength;
    var dataType;
    var editLength;

    /* 各値をダブルクリックした際に編集エリアに反映 */
    $(document).on("dblclick", ".sfo-value", function(){
        keyName = $(this).parent().attr("data-key");
		dataValue = $(this).parent().find(".sfo-value").text();
        dataLength = $(this).parent().find(".nume").text();
        dataMaxLength = $(this).parent().find(".deno").text();
        dataType = $(this).parent().find(".sfo-type").text();
        $("#edit-key").val(keyName);
        $("#edit-value").val(dataValue)
        $("#edit-length").val("Length (Byte) - " + dataLength + "/" + dataMaxLength);
        if(dataType==PSF_TYPE_UTF8S || dataType==PSF_TYPE_INT32){
            $("#edit-value").addClass("input-num-only");
        }else{
            $("#edit-value").removeClass("input-num-only")
        }
        $("#edit-length").removeClass("edit-length-excess");
	})

    /* input-num-onlyクラスが設定されている場合の処理 */
    $(document).on("keydown", ".input-num-only", function(e){
        var k = e.keyCode;
        var kAry = [8, 46, 17, 86, 16, 65, 88, 90, 35, 36];
        var kAry2 = [66, 67, 68, 69, 70] 
        var str = String.fromCharCode(k);
        // 数字、バックスペース、Delet、矢印、コピペ、切り取り、Home、End以外は無効化
        // タイプが0ならB～FもOK
        if(dataType==PSF_TYPE_INT32){
            if(!(str.match(/[0-9]/) || (37<=k && k<=40) || kAry.indexOf(k)>-1)){
                return false;
            }
        }else if(dataType==PSF_TYPE_UTF8S){
            kAry = kAry.concat(kAry2);
            if(!(str.match(/[0-9]/) || (37<=k && k<=40) || kAry.indexOf(k)>-1)){
                return false;
            }
        }
    })
    $(document).on("keyup blur", ".input-num-only", function(){
        if(dataType==PSF_TYPE_INT32){
            this.value = this.value.replace(/[^0-9]+/i,"");
        }else if(dataType==PSF_TYPE_UTF8S){
            this.value = this.value.replace(/[^0-9a-f]+/i,"");
            var editValue = $("#edit-value").val();
            editLength = hexStringToBytes(editValue).length;
            if(editLength > dataMaxLength){
                $("#edit-length").addClass("edit-length-excess");
            }else{
                $("#edit-length").removeClass("edit-length-excess");
            }
            $("#edit-length").val( "Length (Byte) - " + editLength + "/" + dataMaxLength);
        }
    })

    /* 入力時に長さを計算 */
    $("#edit-value").on("input", function(){
        var editValue = $("#edit-value").val();
        if(dataType==PSF_TYPE_INT32){
            editLength = 4;
            if(editValue > 4294967295){
                $("#edit-value").val(dataValue);
            }
        }else if(dataType==PSF_TYPE_UTF8){
            editLength = stringToBytes(editValue).length + 1;
        }else{
            editLength = hexStringToBytes(editValue).length;
        }
        if(editLength > dataMaxLength){
            $("#edit-length").addClass("edit-length-excess");
        }else{
            $("#edit-length").removeClass("edit-length-excess");
        }
        $("#edit-length").val( "Length (Byte) - " + editLength + "/" + dataMaxLength);
    })

    /* Changeボタンを押した際の処理 */
    $("#edit-change").on("click", function(){
        var editValue = $("#edit-value").val();
        // 長さの計算
        if(dataType==PSF_TYPE_INT32){
            editLength = 4;
        }else if(dataType==PSF_TYPE_UTF8){
            editValue = editValue + String.fromCodePoint(0);
            editLength = stringToBytes(editValue).length;
        }else{
            editLength = hexStringToBytes(editValue).length;
        }
        // オブジェクトに保存し、テーブルに反映
        if(editLength <= dataMaxLength){
            var dataPadding = dataMaxLength - editLength;
            editEntries[keyName]["value"] = editValue;
            editEntries[keyName]["length"] = editLength;
            editEntries[keyName]["valpadding"] = dataPadding;
            var trElem = "[data-key='" + keyName + "']";
            $(trElem).find(".sfo-value").html(editValue)
            $(trElem).find(".sfo-length").find(".nume").html(editLength)
        }else{
            alert("Value is too long!")
        }
    })

    /* Resetボタンを押した際に値をリセット */
    $("#edit-reset").on("click", function(){
        if(dataValue!=undefined){
            $("#edit-value").val(dataValue);
            $("#edit-length").val("Length (Byte) - " + dataLength + "/" + dataMaxLength);
        }
    })

    /* Saveボタンを押した際の処理 */
    $("#save-btn").off("click");
    $("#save-btn").on("click", function(){
       const sfo = createSfo(editEntries);
        var a = document.createElement("a");
        document.body.appendChild(a);
        a.style = "display: none";
        var blob = new Blob([sfo], {type: "octet/stream"}),
        url = window.URL.createObjectURL(blob);
        a.href = url;
        sfoFileName = $("#edit-file-name").val();
        a.download = sfoFileName;
        a.click();
        window.URL.revokeObjectURL(url);
        $(a).remove();
    })
}

function addParams(){
    var ps3ParamsObj = {};
    var ps4ParamsObj = {};
    var pspParamsObj = {};
    var psvParamsObj = {};
    var systemName = "ps3";
    var addKeyName = "ACCOUNT_ID";
    var addMaxLength = 16;
    var addLength = 0;
    var addType = 0;
    
    $("#add-params-btn1").on("click", function(){
        if(modalNotCreated){
            // 初めてAddボタンを押した場合はセレクトボックスとオブジェクトを作成
            modalNotCreated = false;
            getXmlData().then(
                function(data){
                    var ps3OptionElemAray = [];
                    var ps4OptionElemAray = [];
                    var pspOptionElemAray = [];
                    var psvOptionElemAray = [];
                    $.each($(data).find("param"), function(i,param){
                        var keyName = $(param).attr("key");
                        var fmt = $(param).attr("fmt");
                        var maxLen = $(param).attr("max_len");
                        var sys = $(param).attr("sys");
                        var optionElem = "<option value='" + keyName + "'>" + keyName + "</option>";
                        switch(sys){
                            case "PS3":
                                ps3ParamsObj[keyName] = {
                                    "fmt": fmt,
                                    "max_len": maxLen,
                                    "sys": sys
                                }
                                ps3OptionElemAray.push(optionElem);
                                break;
                            case "PS4":
                                ps4ParamsObj[keyName] = {
                                    "fmt": fmt,
                                    "max_len": maxLen,
                                    "sys": sys
                                }
                                ps4OptionElemAray.push(optionElem);
                                break;
                            case "PSP":
                                pspParamsObj[keyName] = {
                                    "fmt": fmt,
                                    "max_len": maxLen,
                                    "sys": sys
                                }
                                pspOptionElemAray.push(optionElem);
                                break;
                            case "PSV":
                                psvParamsObj[keyName] = {
                                    "fmt": fmt,
                                    "max_len": maxLen,
                                    "sys": sys
                                }
                                psvOptionElemAray.push(optionElem);
                                break
                        }
                    })
                    var ps3ParamSb = "<select id='ps3-param-sb'>"  + ps3OptionElemAray.join("") + "</select>";
                    var ps4ParamSb = "<select id='ps4-param-sb'>"  + ps4OptionElemAray.join("") + "</select>";
                    var pspParamSb = "<select id='psp-param-sb'>"  + pspOptionElemAray.join("") + "</select>";
                    var psvParamSb = "<select id='psv-param-sb'>"  + psvOptionElemAray.join("") + "</select>";
                    $(".modal-key-wrapper").append(ps3ParamSb);
                    $(".modal-key-wrapper").append(ps4ParamSb);
                    $(".modal-key-wrapper").append(pspParamSb);
                    $(".modal-key-wrapper").append(psvParamSb);
                    $("#ps4-param-sb").hide();
                    $("#psp-param-sb").hide();
                    $("#psv-param-sb").hide();
                    $("#add-length").val( "Length (Byte) - " + 0 + "/" + addMaxLength);
                    //  Systemラジオボタンを変更した場合の処理
                    $("input[name='systemRadioOptions']").on("click", function(){
                        systemName = $(this).val();
                        switch(systemName){
                            case "ps3":
                                $("#ps3-param-sb").show();
                                $("#ps4-param-sb").hide();
                                $("#psp-param-sb").hide();
                                $("#psv-param-sb").hide();
                                break;
                            case "ps4":
                                $("#ps3-param-sb").hide();
                                $("#ps4-param-sb").show();
                                $("#psp-param-sb").hide();
                                $("#psv-param-sb").hide();
                                break;
                            case "psp":
                                $("#ps3-param-sb").hide();
                                $("#ps4-param-sb").hide();
                                $("#psp-param-sb").show();
                                $("#psv-param-sb").hide();
                                break;
                            case "psv":
                                $("#ps3-param-sb").hide();
                                $("#ps4-param-sb").hide();
                                $("#psp-param-sb").hide();
                                $("#psv-param-sb").show();
                                break;
                        }
                    })
                    // Keyを選択した場合の処理
                    $(".modal-key-wrapper select").on("change", function(){
                        addKeyName = $(this).find("option:selected").val();
                        var addFmt;
                        switch(systemName){
                            case "ps3":
                                addMaxLength = ps3ParamsObj[addKeyName]["max_len"];
                                addFmt = ps3ParamsObj[addKeyName]["fmt"];
                                break;
                            case "ps4":
                                addMaxLength = ps4ParamsObj[addKeyName]["max_len"];
                                addFmt = ps4ParamsObj[addKeyName]["fmt"];
                                break;
                            case "psp":
                                addMaxLength = pspParamsObj[addKeyName]["max_len"];
                                addFmt = pspParamsObj[addKeyName]["fmt"];
                                break;
                            case "psv":
                                addMaxLength = psvParamsObj[addKeyName]["max_len"];
                                addFmt = psvParamsObj[addKeyName]["fmt"];
                                break;
                        }
                        addMaxLength = parseInt(addMaxLength);
                        switch(addFmt){
                            case "int32":
                                addType = PSF_TYPE_INT32;
                                $("input[name='typeRadioOptions']"+"[value='"+addType+"']").prop("checked", true);
                                break;
                            case "utf8":
                                addType = PSF_TYPE_UTF8;
                                $("input[name='typeRadioOptions']"+"[value='"+addType+"']").prop("checked", true);
                                break;
                            default:
                                addType = PSF_TYPE_UTF8S;
                                break;
                        }
                        addType = parseInt(addType);
                        $("#add-value").val("");;
                        addLength = 0;
                        $("#add-length").val( "Length (Byte) - " + addLength + "/" + addMaxLength);
                    })
                    // Typeラジオボタンを変更した場合の処理
                    $("input[name='typeRadioOptions']").on("click", function(){
                        addType = $(this).val();
                        addType = parseInt(addType);
                        if(addType==PSF_TYPE_INT32){
                            addMaxLength = 4;
                        }else{
                            switch(systemName){
                                case "ps3":
                                    addMaxLength = ps3ParamsObj[addKeyName]["max_len"];
                                    break;
                                case "ps4":
                                    addMaxLength = ps4ParamsObj[addKeyName]["max_len"];
                                    break;
                                case "psp":
                                    addMaxLength = pspParamsObj[addKeyName]["max_len"];
                                    break;
                                case "psv":
                                    addMaxLength = psvParamsObj[addKeyName]["max_len"];
                                    break;
                            }
                        }
                        checkLength();
                    })
                }
            )
        }else{
            $("#add-length").val( "Length (Byte) - " + 0 + "/" + addMaxLength);
        }
    })

    // editSfo()で定義したのとほぼ同じ
    $("#add-value").on("keydown", function(e){
        var k = e.keyCode;
        var kAry = [8, 46, 17, 86, 16, 65, 88, 90, 35, 36];
        var kAry2 = [66, 67, 68, 69, 70] 
        var str = String.fromCharCode(k);
        if(addType==PSF_TYPE_INT32){
            if(!(str.match(/[0-9]/) || (37<=k && k<=40) || kAry.indexOf(k)>-1)){
                return false;
            }
        }else if(addType==PSF_TYPE_UTF8S){
            kAry = kAry.concat(kAry2);
            if(!(str.match(/[0-9]/) || (37<=k && k<=40) || kAry.indexOf(k)>-1)){
                return false;
            }
        }
    })
    $("#add-value").on("keyup blur", function(){
        if(addType==PSF_TYPE_INT32){
            this.value = this.value.replace(/[^0-9]+/i,"");
        }else if(addType==PSF_TYPE_UTF8S){
            this.value = this.value.replace(/[^0-9a-f]+/i,"");
            var addValue = $("#add-value").val();
            addLength = hexStringToBytes(addValue).length;
            if(addLength > addMaxLength){
                $("#add-length").addClass("edit-length-excess");
            }else{
                $("#add-length").removeClass("edit-length-excess");
            }
            $("#add-length").val( "Length (Byte) - " + addLength + "/" + addMaxLength);
        }
    })
    $("#add-value").on("input", function(){
        checkLength();
    })

    // モーダル内のAddボタンを押した際の処理
    $("#add-params-btn2").off("click");
    $("#add-params-btn2").on("click", function(){
        var addValue = $("#add-value").val();
        if(addType==PSF_TYPE_INT32){
            addLength = 4;
        }else if(addType==PSF_TYPE_UTF8){
            addValue = addValue + String.fromCodePoint(0);
            addLength = stringToBytes(addValue).length;
        }else if(addType==PSF_TYPE_UTF8S){
            addLength = hexStringToBytes(addValue).length
        }
        console.log(addValue);
        if(addLength <= addMaxLength){
            var dataPadding = addMaxLength - addLength;
            editEntries[addKeyName] = {
                "value": addValue,
                "length": addLength,
                "maxLength": addMaxLength,
                "type": addType,
                "valpadding": dataPadding
            }
            // ソート
            var sfoKeyNames = Object.keys(editEntries);
            sfoKeyNames.sort();
            var sortedEntries = {};
            for(var i=0; i<sfoKeyNames.length; i++){
                sortedEntries[sfoKeyNames[i]] = editEntries[sfoKeyNames[i]]
            }
            editEntries = {};
            editEntries = $.extend(true, {}, sortedEntries);
            $("#sfo-table").find("tbody").empty();
            sfoFileName = $("#edit-file-name").val();
            displaySfo(editEntries, sfoFileName, $("#sfo-table"))
            $("#add-value").val("");
            $("#add-length").val("");
            $("#addParamsModal").modal("hide");
        }else{
            alert("Value is too long!")
        }
    })

    // xmlの読み取り
    function getXmlData(){
        var def = $.Deferred();
        $.ajax({
            url: "SFOEditor.xml",
            type: "GET",
            dataType: "xml"
        }).done(function(data){
            def.resolve(data);
        });
        return def.promise();
    }

    function checkLength(){
        var addValue = $("#add-value").val();
        if(addType==PSF_TYPE_INT32){
            addLength = 4;
            if(addValue > 4294967295){
                $("#add-value").val(0);
            }
        }else if(addType==PSF_TYPE_UTF8){
            addLength = stringToBytes(addValue).length + 1;
        }else{
            addLength = hexStringToBytes(addValue).length;
        }
        if(addLength > addMaxLength){
            $("#add-length").addClass("edit-length-excess");
        }else{
            $("#add-length").removeClass("edit-length-excess");
        }
        $("#add-length").val( "Length (Byte) - " + addLength + "/" + addMaxLength);
    }

}